"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.swaggerToTypeScript = void 0;
const lodash_1 = __importDefault(require("lodash"));
const prettier_1 = __importDefault(require("prettier"));
const swagger2openapi_1 = require("swagger2openapi");
const parse_1 = require("./parse");
const render_1 = require("./render");
async function swaggerToTypeScript(document, options) {
    const parser = new parse_1.Parser();
    if (document.swagger?.startsWith('2.')) {
        fixMissingRef(document);
        await parser.init((await (0, swagger2openapi_1.convertObj)(document, { resolve: false })).openapi);
    }
    else {
        await parser.init(document);
    }
    const code = [
        ` /* eslint-disable */
      import { toFormData } from 'axios'
      import type { AxiosResponse, AxiosRequestConfig, AxiosInstance } from 'axios'
      type PickData<T> = T extends { data?: any } ? T['data'] : any
      const interpolatePath = (str: string, values: any) => str.replace(/{([^{}]+)}/g, (_, g1) => values[g1])

      ${renderTypeConversion(options?.openAPITypes)}
    `,
    ];
    const ops = parser.getProcessedOperationObjects() || [];
    const tasks = ops.map((el) => () => (0, render_1.renderOperation)(el));
    while (tasks.length) {
        code.push(await tasks.pop()());
    }
    code.push(`
    const collectApis = () => ({
      ${ops
        .reduce((acc, el) => [...acc, `...${el.operationId}(),`], [])
        .join('\n')}
    });

    export type Collection = ReturnType<typeof collectApis>
    export type Keys = keyof Collection
    export type ApiTypes<K extends Keys> = Collection[K][0]
    export type ApiRequestParams<K extends Keys> = Parameters<Collection[K][1]>
    
    export type ApiDataTypes<K extends Keys, T extends keyof ApiTypes<Keys>> = ApiTypes<K>[T]
    
    export function createRequest(axios: AxiosInstance) {
      const collections = collectApis()
    
      return function request<K extends Keys, U = any>(key: K) {
        const fn = collections[key][1] as (...args: any[]) => AxiosRequestConfig;
    
        return (...args: ApiRequestParams<K>) => axios(fn(...args)) as Promise<U>
      }
    }
    
    type Arg<K extends Keys> = ApiRequestParams<K>
    type BaseResp<K extends Keys, U extends "Response" | "ResponseData"> = Promise<AxiosResponse<ApiDataTypes<K, U>>>;

    export interface BaseRequest<T extends "Response" | "ResponseData"> {
      ${ops
        .map((el) => ({
        ...el,
        title: el.description || '',
        req: `"${el.__extra__.method.toUpperCase()} ${el.__extra__.url}"`,
    }))
        .map((el) => `
          /** @summary ${el.summary} */
          <K extends ${el.req}>(key: K): (...args: Arg<K>) => BaseResp<K, T>;
        `)
        .join('\n')}
    }

    export function createAxiosAPI(axios: AxiosInstance) {
      return createRequest(axios) as BaseRequest<'Response'>
    }
  `);
    const fullCode = await prettier_1.default.format(code.join('\n'), {
        parser: 'typescript',
        tabWidth: 2,
        printWidth: 120,
        proseWrap: 'never',
        singleQuote: true,
        semi: false,
        trailingComma: 'es5',
        endOfLine: 'auto',
    });
    return fullCode;
}
exports.swaggerToTypeScript = swaggerToTypeScript;
function fixMissingRef(obj, root = obj) {
    lodash_1.default.forEach(obj, (value) => {
        if (value !== null && typeof value === 'object') {
            // 删除无效引用
            if ('$ref' in value) {
                if (!lodash_1.default.get(root, value.$ref.replace('#/', '').split('/'))) {
                    delete value.$ref;
                    delete value.originalRef;
                }
            }
            else {
                lodash_1.default.forEach(value, (val) => fixMissingRef(val, root));
            }
        }
    });
    return obj;
}
function renderTypeConversion(opt) {
    const getBooleanType = (key, defaults = 'boolean') => lodash_1.default.get(opt, ['boolean', key], defaults);
    const getNumberType = (key, defaults = 'number') => lodash_1.default.get(opt, ['number', key], defaults);
    const getIntType = (key, defaults = 'number') => lodash_1.default.get(opt, ['integer', key], defaults);
    const getStringType = (key, defaults = 'string') => lodash_1.default.get(opt, ['string', key], defaults);
    const stringFormats = lodash_1.default.without(lodash_1.default.keys(opt?.string), ...['default', 'date', 'date-time', 'password', 'binary', 'byte']);
    return `
      type $$Boolean = {
        /** @SEE https://swagger.io/docs/specification/data-models/data-types/#boolean */
        ['default']: ${getBooleanType('default')}
      }

      type $$Number = {
        /**
         * Any numbers.
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#numbers
         */
        ['default']: ${getNumberType('default')}
        /**
         * Floating-point numbers.
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#numbers
         */
        ['float']: ${getNumberType('float')}
        /**
         * Floating-point numbers with double precision.
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#numbers
         */
        ['double']: ${getNumberType('double')}
      }

      type $$Integer = {
        /**
         * Integer numbers.
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#numbers
         */
        ['default']: ${getIntType('default')}
        /**
         * Signed 32-bit integers (commonly used integer type).
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#numbers
         */
        ['int32']: ${getIntType('int32')}
        /**
         * Signed 64-bit integers (long type).
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#numbers
         */
        ['int64']: ${getIntType('int64', 'string')}
      }

      // https://swagger.io/docs/specification/data-models/data-types/#string
      type $String = {
        /**
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#string
         */
        ['default']: ${getStringType('default')}
        /**
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#string
         */
        ['date']: ${getStringType('date')}
        /**
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#string
         */
        ['date-time']: ${getStringType('date-time')}
        /**
         * a hint to UIs to mask the input
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#string
         */
        ['password']: ${getStringType('password')}
        /**
         * binary file contents
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#file
         */
        ['binary']: ${getStringType('binary', 'File')}
        /**
         * 通常代表 base64-encoded file contents
         * 但是实际使用中貌似也有例外
         * @SEE https://swagger.io/docs/specification/data-models/data-types/#file
         */
        ['byte']: ${getStringType('byte')}

        ${stringFormats.map((key) => `['${key}']: ${getStringType(key)};`).join('\n')}
      }
    

      type $Get<T extends { default: any; [k: string]: any }, K extends string> = [K] extends [
        keyof T,
      ]
        ? T[K]
        : T['default']

      type __GetType<
        // https://cswr.github.io/JsonSchema/spec/basic_types/
        // https://swagger.io/docs/specification/data-models/data-types/
        Type extends 'string' | 'number' | 'integer' | 'boolean',
        Fmt extends string = 'default',
      > = [Type] extends ['number']
        ? $Get<$$Number, Fmt>
        : [Type] extends ['integer']
        ? $Get<$$Integer, Fmt>
        : [Type] extends ['boolean']
        ? $Get<$$Boolean, Fmt>
        : [Type] extends ['string']
        ? $Get<$String, Fmt>
        : never
        
  `;
}
//# sourceMappingURL=index.js.map