"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Parser = void 0;
const lodash_1 = __importDefault(require("lodash"));
const openapi_types_1 = require("openapi-types");
const swagger_parser_1 = __importDefault(require("@apidevtools/swagger-parser"));
const HTTP_METHODS = Object.values(openapi_types_1.OpenAPIV3.HttpMethods);
function resolveCircular(obj, seen) {
    lodash_1.default.forEach(obj, (value, key) => {
        if (value !== null && typeof value === 'object') {
            // 简单解决循环引用
            if (seen.has(value)) {
                if (key === 'items') {
                    obj[key] = { tsType: '/** recursive */ any' };
                }
                else if (key === 'properties') {
                    obj[key] = lodash_1.default.mapValues(value, (v) => {
                        return { tsType: '/** recursive */ any' };
                    });
                }
                else {
                    return {};
                }
            }
            else {
                seen.add(value);
                lodash_1.default.forEach(value, (val) => resolveCircular(val, seen));
                seen.delete(value);
            }
        }
    });
    return obj;
}
class Parser {
    constructor() {
        this.document = null;
        this.parser = new swagger_parser_1.default();
    }
    async init(api) {
        const doc = await swagger_parser_1.default.dereference(api);
        this.document = resolveCircular(doc, new WeakSet());
    }
    getProcessedOperationObjects() {
        if (!this.document)
            return null;
        const { paths } = this.document;
        const pathKeys = Object.keys(paths);
        return pathKeys
            .filter((key) => paths[key])
            .map((key) => [key, paths[key]])
            .map(([key, path]) => HTTP_METHODS.filter((method) => path?.[method]).map((method) => [key, method, { ...path[method] }]))
            .flat(1)
            .map(([url, method, item]) => this.processOperationObject(item, { url, method }));
    }
    processOperationObject(item, { url, method }) {
        const [requestBody, isMultipart] = (() => {
            const bodyContent = item.requestBody?.content;
            const mediaObject = bodyContent
                ? bodyContent?.['application/json'] ||
                    bodyContent?.['multipart/form-data'] ||
                    lodash_1.default.values(bodyContent)[0]
                : null;
            const isMultipart = Boolean(bodyContent?.['multipart/form-data']);
            return [mediaObject, isMultipart];
        })();
        const parameters = this.processParametersObject(item.parameters) || [];
        const pathParams = parameters
            .filter((el) => el.in === 'path')
            .sort((a, b) => Number(b.schema?.required || 0) - Number(a.schema?.required || 0));
        const queryParams = parameters.filter((el) => el.in === 'query');
        const responseData = lodash_1.default.values(item.responses['200']?.content)?.[0];
        return {
            ...item,
            isMultipart,
            parameters: {
                pathParams,
                queryParams,
            },
            responseData,
            requestBody: requestBody,
            __extra__: { url, method },
        };
    }
    processParametersObject(parameters) {
        return parameters
            ?.map((el) => el)
            .sort((a, b) => a.in.localeCompare(b.in))
            .map((param) => ({
            ...param,
            schema: param.schema,
        }));
    }
}
exports.Parser = Parser;
//# sourceMappingURL=parse.js.map